/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Copyright (C) 2005, Oxymium sarl
 * 
 * Manuel Guesdon <mguesdon@oxymium.net> - LDAP RealTime Driver Author/Adaptor
 *
 * res_config_ldap.c <LDAP plugin for RealTime configuration engine>
 *
 * v0.5   - (11-16-2005) - Initial version based
 */

#include <asterisk/channel.h>
#include <asterisk/logger.h>
#include <asterisk/config.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <asterisk/options.h>
#include <asterisk/cli.h>
#include <asterisk/utils.h>
#include <stdlib.h>
#include <string.h>
#include <ldap.h>

static char *res_config_ldap_desc = "LDAP RealTime Configuration Driver";
AST_MUTEX_DEFINE_STATIC(ldap_lock);
#define RES_CONFIG_LDAP_CONF "res_ldap.conf"
static LDAP*  ldapConn = NULL;
static char   dbhost[512]="";
static char   dbuser[512]="";
static char   dbpass[50]="";
static char   dbbasedn[512]="";
static int    dbport=389;
static time_t connect_time=0;

static int parse_config(void);
static int ldap_reconnect(void);
static int realtime_ldap_status(int fd, int argc, char **argv);


STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static char cli_realtime_ldap_status_usage[] =
"Usage: realtime ldap status\n"
"       Shows connection information for the LDAP RealTime driver\n";

static struct ast_cli_entry cli_realtime_ldap_status = {
        { "realtime", "ldap", "status", NULL }, realtime_ldap_status,
        "Shows connection information for the LDAP RealTime driver", cli_realtime_ldap_status_usage, NULL };

// Log variable
static void ast_variable_log(const char* log_prefix,struct ast_variable* var)
{
  ast_log(LOG_DEBUG, "LOG Variable: %s\n",log_prefix);
  while(var)
    {
      ast_log(LOG_DEBUG, "%s => %s\n",var->name,var->value);
      var=var->next;
    };
  ast_log(LOG_DEBUG, "END LOG Variable: %s\n",log_prefix);
}

// Log config
static void ast_config_log(const char* log_prefix,struct ast_config* config)
{
  ast_log(LOG_DEBUG, "LOG Config: %s\n",log_prefix);
  char* category_name=NULL;
  while((category_name=ast_category_browse(config,category_name)))
    {
      ast_log(LOG_DEBUG, "========= category: %s ==========\n",category_name);
      struct ast_variable* var=ast_variable_browse(config,category_name);
      if (var)
	{
	  while(var)
	    {
	      ast_log(LOG_DEBUG, "%s=%s\n",var->name,var->value);
	      var = var->next;
	    };
	};
    }
  ast_log(LOG_DEBUG, "END LOG Config: %s\n",log_prefix);
};

// Table configuration
struct ldap_table_config
{
  char* table_name;			// table name
  char* additional_filter;		// additional filter       
  struct ast_variable* attributes;	// attribute names conversion
  struct ldap_table_config* next;	// next table
};

// Should be locked before using it
static struct ldap_table_config* table_configs=NULL;

// Create a new table_config
static struct ldap_table_config* table_config_new(const char* table_name)
{
  struct ldap_table_config* p=malloc(sizeof(struct ldap_table_config));
  memset(p,0,sizeof(struct ldap_table_config));
  if (table_name)
    p->table_name=strdup(table_name);
  return p;
};

// Find a table_config - Should be locked before using it
static struct ldap_table_config* table_config_for_table_name(const char* table_name)
{
  struct ldap_table_config* c=table_configs;
  while(c)
    {
      if (strcmp(c->table_name,table_name)==0)
	return c;
      else
	c=c->next;
    };
  return NULL;
};

// add attributes to table config - Should be locked before using it
static void ldap_table_config_add_attributes(struct ldap_table_config* table_config,const char* attributes_string)
{
  if (attributes_string && *attributes_string)
    {      
      char* string=strdup(attributes_string);
      char* string_end=string+strlen(string);
      char* start=string;
      //ast_log(LOG_DEBUG, "LDAP RealTime: Add attribute: start: %s\n",start);
      while(start<string_end)
	{
	  char* conv_sep=NULL;
	  //ast_log(LOG_DEBUG, "LDAP RealTime: Add attribute: start: %s\n",start);
	  char* attr_sep=strchr(start,',');
	  //ast_log(LOG_DEBUG, "LDAP RealTime: Add attribute: attr_sep: %s\n",attr_sep);
	  if (!attr_sep)
	    attr_sep=string_end;
	  else
	    *attr_sep='\0';
	  conv_sep=strchr(start,'=');
	  //ast_log(LOG_DEBUG, "LDAP RealTime: Add attribute: conv_sep: %s\n",conv_sep);
	  if (conv_sep>=attr_sep)
	    conv_sep=NULL;

	  if (!conv_sep)
	    {
	    }
	  else
	    {
	      *conv_sep='\0';
	      struct ast_variable* var=ast_variable_new(start,conv_sep+1);
	      //ast_log(LOG_DEBUG, "LDAP RealTime: Add attribute: VAR %s => %s\n",var->name,var->value);
	      if (table_config->attributes)
		var->next=table_config->attributes;
	      table_config->attributes=var;
	    }
	  start=attr_sep+1;
	};
      free(string);
    };
};
				    

// Free table_config
static void table_configs_free(void)
{
  struct ldap_table_config* c=table_configs;
  while(c)
    {
      struct ldap_table_config* next=c->next;
      if (c->table_name)
	free(c->table_name);
      if (c->additional_filter)
	free(c->additional_filter);
      if (c->attributes)
	{
	  ast_variables_destroy(c->attributes);
	};
      free(c);
      c=next;
    };
}

// Convert variable name to ldap attribute name - Should be locked before using it
static const char* convert_attribute_name_to_ldap(struct ldap_table_config* table_config,const char* attribute_name)
{
  if (table_config)
    {
      struct ast_variable* attribute=table_config->attributes;
      while(attribute)
	{
	  if (strcasecmp(attribute_name,attribute->name)==0)
	    return attribute->value;
	  else
	    attribute=attribute->next;
	};
    };
  return attribute_name;
};

// Convert ldap attribute name to variable name - Should be locked before using it
static const char* convert_attribute_name_from_ldap(struct ldap_table_config* table_config,const char* attribute_name)
{
  if (table_config)
    {
      struct ast_variable* attribute=table_config->attributes;
      while(attribute)
	{
	  if (strcasecmp(attribute_name,attribute->value)==0)
	    return attribute->name;
	  else
	    attribute=attribute->next;
	};
    };
  return attribute_name;
};

// Find variable by name
static struct ast_variable *variable_named(struct ast_variable *var,const char* name)
{
  while(var)
    {
      if (strcasecmp(name,var->name)==0)
	return var;
      else
	var=var->next;
    }
  return NULL;
}

// Get variables from ldap entry attributes - Should be locked before using it
static struct ast_variable *realtime_ldap_entry_to_var(struct ldap_table_config* table_config, LDAPMessage *ldap_entry)
{
  BerElement *ber=NULL;
  struct ast_variable *var=NULL;
  struct ast_variable *prev=NULL;
  
  char* ldap_attribute_name = ldap_first_attribute(ldapConn,ldap_entry,&ber);
  
  while(ldap_attribute_name) {
    const char* attribute_name=convert_attribute_name_from_ldap(table_config,ldap_attribute_name);
    int is_realmed_password_attribute=strcasecmp(attribute_name,"md5secret")==0;
    char **values=NULL;
    
    values = ldap_get_values(ldapConn,ldap_entry,ldap_attribute_name);
    if (!values)
      {
      }
    else
      {
	char **v=values;
	
	while(*v) {
	  char* value=*v;
	  ast_log(LOG_DEBUG, "LDAP RealTime: attribute_name: %s value: %s\n", attribute_name,value);
	  if (is_realmed_password_attribute)
	    {
	      if (strncasecmp(value,"{md5}",5)==0)
		value+=5;
	      else
		value=NULL;
	    };
	  if (value) {
	    if(prev) {
	      prev->next = ast_variable_new(attribute_name, value);
	      if (prev->next) {
		prev = prev->next;
	      }
	    }
	    else {
	      prev = var = ast_variable_new(attribute_name, value);
	    }
	  }
	  v++;
	};
	ldap_value_free(values);
      }
    ldap_attribute_name =ldap_next_attribute(ldapConn,ldap_entry,ber);
  }
  ber_free(ber,0);

  return var;
}

// LGet LDAP entry by dn and return attributes as variables  - Should be locked before using it
static struct ast_variable *ldap_loadentry(struct ldap_table_config* table_config, const char *dn)
{
  struct ast_variable *var=NULL;
  ast_log(LOG_DEBUG, "LDAP RealTime: ldap_loadentry dn=%s\n", dn);

  int result = 0;
  LDAPMessage *ldap_result=NULL;
  result=ldap_search_s(ldapConn,dn,LDAP_SCOPE_BASE,
		       table_config->additional_filter,NULL,0,&ldap_result);
  if (result<0) {
    ast_log(LOG_WARNING, "LDAP RealTime: Failed to query database. Check debug for more info.\n");
    ast_log(LOG_DEBUG, "LDAP RealTime: dn=%s\n", dn);
    ast_log(LOG_DEBUG, "LDAP RealTime: Query Failed because: %s\n", ldap_err2string(result));
    ast_mutex_unlock(&ldap_lock);
    return NULL;
  }
  else {
    int num_entry = 0;
    if((num_entry=ldap_count_entries(ldapConn,ldap_result))>0) {
      LDAPMessage *ldap_entry=NULL;
      ast_log(LOG_DEBUG, "LDAP RealTime: num_entry: %d\n", num_entry);
      
      ldap_entry = ldap_first_entry(ldapConn,ldap_result);
      var=realtime_ldap_entry_to_var(table_config, ldap_entry);
      if (num_entry>1) {
	  ast_log(LOG_WARNING, "LDAP RealTime: More than one entry for dn=%s. Take only 1st one\n", dn);
      }
    }
    else {
      ast_log(LOG_WARNING, "LDAP RealTime: Could not find any entry dn=%s.\n", dn);
    }
  }    
  ldap_msgfree(ldap_result);
  
  ast_variable_log("LDAP RealTime: Found Entry",var);
  return var;
}

// Append a string to a filter string. The filter string can grow
static void append_string_to_filter(char** filter_ptr,int* filter_size_ptr,const char* filter)
{
  int current_len=(*filter_ptr ? strlen(*filter_ptr) : 0);
  int needed_len=current_len+strlen(filter);
  if (*filter_size_ptr<(needed_len+1))
    {
      if (*filter_size_ptr==0)
	*filter_size_ptr=(needed_len>128 ? needed_len : 128);
      else
	*filter_size_ptr=(*filter_size_ptr)*2;
      *filter_ptr=realloc(*filter_ptr,*filter_size_ptr);
    }
  if (*filter_ptr)
    {
      strcpy((*filter_ptr)+current_len,filter);
    }
}

// Replace search by by in string. No check is done on string allocated size !
static int replace_string_in_string(char* string,const char* search,const char* by)
{
  int search_len=strlen(search);
  int by_len=strlen(by);
  int replaced=0;
  char* p=strstr(string,search);
  if (p)
    {
      replaced=1;
      while(p)
	{
	  if (by_len==search_len)
	    memcpy(p,by,by_len);
	  else
	    {
	      memmove(p+by_len,p+search_len,strlen(p+search_len)+1);
	      memcpy(p,by,by_len);
	    }
	  p=strstr(p+by_len,search);
	};
    };
  return replaced;
};

// Append a name=value filter string. The filter string can grow.
// convert name and value if "LIKE' is used (see http://bugs.digium.com/view.php?id=5765)
static void append_var_and_value_to_filter(char** filter_ptr,int* filter_size_ptr,
					   struct ldap_table_config* table_config,
					   const char* name,const char* value)
{  
  char* new_name=NULL;
  char* new_value=NULL;
  char* wsPos=strstr(name," LIKE");
  //ast_log(LOG_DEBUG, "LDAP RealTime: name='%s' value='%s'\n",name,value);
  if (wsPos)
    {      
      new_name=strndup(name,wsPos-name);
      name=new_name;
      new_value=strdup(value);
      replace_string_in_string(new_value,"\\_","_");
      replace_string_in_string(new_value,"%","*");
      value=new_value;
    }
  name=convert_attribute_name_to_ldap(table_config,name);
  append_string_to_filter(filter_ptr,filter_size_ptr,"(");
  append_string_to_filter(filter_ptr,filter_size_ptr,name);
  append_string_to_filter(filter_ptr,filter_size_ptr,"=");
  append_string_to_filter(filter_ptr,filter_size_ptr,value);
  append_string_to_filter(filter_ptr,filter_size_ptr,")");
  if (new_name)
    free(new_name);
  if (new_value)
    free(new_value);
}

// LDAP base function
// return a null terminated array of ast_variable (one per entry) or NULL if no entry is found or if an error occured
// caller should free the returned array and ast_variables
// entries_count_ptr is a pointer to found entries count (can be NULL)
// basedn is the base DN
// table_name is the table_name (used dor attribute convertion and additional filter)
// ap contains null terminated list of pairs name/value
static struct ast_variable **realtime_ldap_base(unsigned int *entries_count_ptr,const char *basedn, const char *table_name, va_list ap)
{
        struct ast_variable **vars=NULL;
	if(!table_name) {
		ast_log(LOG_WARNING, "LDAP RealTime: No table_name specified.\n");
	}
	else
	  {
	    const char *newparam=NULL;
	    const char *newval=NULL;

	    // Get the first parameter and first value in our list of passed paramater/value pairs 
	    newparam = va_arg(ap, const char *);
	    newval = va_arg(ap, const char *);
	    if(!newparam || !newval)  {
	      ast_log(LOG_WARNING, "LDAP RealTime: Realtime retrieval requires at least 1 parameter and 1 value to search on.\n");
	    }
	    else
	      {
		ast_mutex_lock(&ldap_lock);

		// We now have our complete statement; Lets connect to the server and execute it. 
		if(!ldap_reconnect()) {
		}
		else {
		  char* filter=NULL;
		  int filter_size=0;
		  struct ldap_table_config* table_config=NULL;
		  
		  int result = 0;
		  LDAPMessage *ldap_result=NULL;
		  		  
		  append_string_to_filter(&filter,&filter_size,"(&");

		  table_config=table_config_for_table_name(table_name);
		  if (table_config && table_config->additional_filter)
		    {
		      append_string_to_filter(&filter,&filter_size,table_config->additional_filter);
		    };
		  
		  // Create the first part of the query using the first parameter/value pairs we just extracted
		  //   If there is only 1 set, then we have our query. Otherwise, loop thru the list and concat

		  append_var_and_value_to_filter(&filter,&filter_size,table_config,newparam, newval);
		  while((newparam = va_arg(ap, const char *))) {
		    newval = va_arg(ap, const char *);
		    append_var_and_value_to_filter(&filter,&filter_size,table_config,newparam, newval);
		  }		  
		  append_string_to_filter(&filter,&filter_size,")");

		  result=ldap_search_s(ldapConn,basedn,LDAP_SCOPE_SUBTREE,
				       filter,NULL,0,&ldap_result);

		  if (result<0) {
		    ast_log(LOG_WARNING, "LDAP RealTime: Failed to query database. Check debug for more info.\n");
		    ast_log(LOG_DEBUG, "LDAP RealTime: Query: %s\n", filter);
		    ast_log(LOG_DEBUG, "LDAP RealTime: Query Failed because: %s\n", ldap_err2string(result));
		  }
		  else {
		    int num_entry = 0;
		    if((num_entry=ldap_count_entries(ldapConn,ldap_result))>0) {
		      LDAPMessage *ldap_entry=NULL;
		      int entry_index=0;
		      if (entries_count_ptr)
			*entries_count_ptr=num_entry;
		      vars=malloc(sizeof(struct ast_variable*)*(num_entry+1));	      
		      memset(vars,0,sizeof(struct ast_variable*)*(num_entry+1));
		      ast_log(LOG_DEBUG, "LDAP RealTime: num_entry: %d\n", num_entry);
		      
		      ldap_entry = ldap_first_entry(ldapConn,ldap_result);
		      
		      for(entry_index=0;ldap_entry;entry_index++) {
			vars[entry_index]=realtime_ldap_entry_to_var(table_config, ldap_entry);
			ast_variable_log("LDAP RealTime: Found Entry",vars[entry_index]);
			ldap_entry = ldap_next_entry(ldapConn,ldap_entry);
			ast_log(LOG_DEBUG, "LDAP RealTime: next entry: %p\n", ldap_entry);
		      };
		    }
		    else {
		      ast_log(LOG_WARNING, "LDAP RealTime: Could not find any entry matching %s in base dn %s.\n", filter, basedn);
		    }
		    
		    ldap_msgfree(ldap_result);

		    if (vars)
		      {
			struct ast_variable **p=vars;
			while(*p) {
			  struct ast_variable *append_var=NULL;
			  struct ast_variable *tmp=*p;
			  while(tmp)
			    {
			      if (strcasecmp(tmp->name,"accountBaseDN")==0)
				{
				  struct ast_variable *base_var=ldap_loadentry(table_config,tmp->value);
				  ast_log(LOG_DEBUG, "LDAP RealTime: Loaded base: %s\n",tmp->value);
				  ast_variable_log("Base Variables",base_var);
				  while(base_var)
				    {
				      struct ast_variable *next=base_var->next;
				      struct ast_variable *test_var=*p;
				      int base_var_found=0;
				      while(test_var)
					{
					  if (strcasecmp(test_var->name,base_var->name)==0)
					    {
					      base_var_found=1;
					      break;
					    }
					  else
					    test_var=test_var->next;
					};
				      if (base_var_found)
					{
					  base_var->next=NULL;
					  ast_variables_destroy(base_var);
					  base_var=next;
					}
				      else
					{
					  if (append_var)
					    base_var->next=append_var;
					  else
					    base_var->next=NULL;
					  append_var=base_var;
					  base_var=next;
					};
				    }
				}
			      if (!tmp->next
				  && append_var)
				{
				  tmp->next=append_var;
				  tmp=NULL;
				}
			      else
				tmp=tmp->next;
			    };
			  p++;
			}		
		      }
		  }
		if (filter)
		  free(filter);
		}
		ast_mutex_unlock(&ldap_lock);
	      }
	  }
	return vars;
}

// same as realtime_ldap_base_ but take variable arguments count list
static struct ast_variable **realtime_ldap_base_(unsigned int *entries_count_ptr,const char *basedn, const char *table_name, ...)
{
  struct ast_variable **vars=NULL;
  va_list ap;
  va_start(ap,table_name);
  vars=realtime_ldap_base(entries_count_ptr,basedn,table_name,ap);
  va_end(ap);
  return vars;
}

// See Asterisk doc
static struct ast_variable *realtime_ldap(const char *basedn, const char *table_name, va_list ap)
{
  struct ast_variable **vars=realtime_ldap_base(NULL,basedn,table_name,ap);
  struct ast_variable *var=NULL;
  if (vars) {
    struct ast_variable *last_var=NULL;
    struct ast_variable **p=vars;
    while(*p) {
      if (last_var) {
	while(last_var->next)
	  last_var=last_var->next;
	last_var->next=*p;
      }
      else {
	var=*p;
	last_var=var;
      }
      p++;
    }
    free(vars);
  }
  ast_variable_log("LDAP RealTime: ==> Found Entry",var);
  return var;
}

// See Asterisk doc
static struct ast_config *realtime_multi_ldap(const char *basedn, const char *table_name, va_list ap)
{
  struct ast_variable **vars=realtime_ldap_base(NULL,basedn,table_name,ap);
  struct ast_config *cfg=NULL;
  if (vars) {
    cfg = ast_config_new();
    if (!cfg) {
      ast_log(LOG_WARNING, "Out of memory!\n");
    }
    else {
      struct ast_variable **p=vars;
      while(*p) {	
	struct ast_category *cat = NULL;
	cat = ast_category_new("");
	if(!cat) {
	  ast_log(LOG_WARNING, "Out of memory!\n");
	  break;
	}
	else {	
	  struct ast_variable *var = *p;
	  while(var) {
	    struct ast_variable *next = var->next;
	    var->next=NULL;
	    ast_variable_append(cat, var);
	    var=next;
	  }
	}
	ast_category_append(cfg, cat);
	p++;
      }
    }
    free(vars);
  }
  return cfg;

}
struct category_and_metric
{
  char* name;
  int metric;
  char* variable_name;
  char* variable_value;
};
static int compare_categories(const void *a, const void *b)
{
  struct category_and_metric* as=(struct category_and_metric*)a;
  struct category_and_metric* bs=(struct category_and_metric*)b;
  if (as->metric<bs->metric)
    return -1;
  else if (as->metric>bs->metric)
    return 1;
  else 
    return strcmp(as->name,bs->name);
}

// See Asterisk doc
static struct ast_config *config_ldap(const char *basedn, const char *table_name, const char *file, struct ast_config *cfg)
{
	if(!file || !strcmp(file, RES_CONFIG_LDAP_CONF)) {
		ast_log(LOG_WARNING, "LDAP RealTime: Cannot configure myself.\n");
	}
	else {
	  unsigned int vars_count=0;
	  struct ast_variable **vars=realtime_ldap_base_(&vars_count,basedn, table_name, "filename", file, "commented", "FALSE", NULL);

	  if (vars) {
	    int i=0;
	    struct ast_variable *new_v;
	    struct ast_category *cur_cat;
	    char* last_category = NULL;
	    int last_category_metric = 0;

	    // sort on metric and category
	    struct category_and_metric* categories=malloc(sizeof(struct category_and_metric)*vars_count);	      
	    struct ast_variable **p=vars;
	    vars_count=0;
	    while(*p) {
	      struct ast_variable* category=variable_named(*p,"category");
	      struct ast_variable* cat_metric=variable_named(*p,"cat_metric");
	      struct ast_variable* var_name=variable_named(*p,"variable_name");
	      struct ast_variable* var_val=variable_named(*p,"variable_value");
	      ast_log(LOG_DEBUG, "LDAP RealTime: category: %s\n", category->value);
	      ast_log(LOG_DEBUG, "LDAP RealTime: var_name: %s\n", var_name->value);
	      ast_log(LOG_DEBUG, "LDAP RealTime: var_val: %s\n", var_val->value);
	      ast_log(LOG_DEBUG, "LDAP RealTime: cat_metric: %s\n", cat_metric->value);
	      if (!category) {
		struct ast_variable* dn=variable_named(*p,"dn");
		ast_log(LOG_ERROR, "LDAP RealTime: No category name in entry '%s'  for file '%s'.\n", 
			(dn ? dn->value : "?"),
			file);
	      }
	      else if (!cat_metric) {
		struct ast_variable* dn=variable_named(*p,"dn");
		ast_log(LOG_ERROR, "LDAP RealTime: No category metric in entry '%s'(category: %s) for file '%s'.\n", 
			(dn ? dn->value : "?"),
			category->value, file);
	      }		
	      else if (!var_name) {
		struct ast_variable* dn=variable_named(*p,"dn");
		ast_log(LOG_ERROR, "LDAP RealTime: No variable name in entry '%s' (category: %s metric: %s) for file '%s'.\n", 
			(dn ? dn->value : "?"), 
			category->value, cat_metric->value, file);
	      }
	      else if (!var_val) {
		struct ast_variable* dn=variable_named(*p,"dn");
		ast_log(LOG_ERROR, "LDAP RealTime: No variable value in entry '%s' (category: %s metric: %s variable: %s) for file '%s'.\n", 
			(dn ? dn->value : "?"),
			category->value, cat_metric->value, var_name->value, file);
	      }	      
	      else {
		categories[vars_count].name=category->value;
		categories[vars_count].metric=atoi(cat_metric->value);
		categories[vars_count].variable_name=var_name->value;
		categories[vars_count].variable_value=var_val->value;
		vars_count++;
	      }
	      p++;
	    }
	    qsort(categories,vars_count,sizeof(struct category_and_metric),compare_categories);
	      
	    for(i=0;i<vars_count;i++) {
	      if(strcmp(categories[i].variable_name, "#include")==0) {
		if (!ast_config_internal_load(categories[i].variable_value, cfg)) {
		  break;
		  return NULL;
		}
	      }
	      else {
		if(!last_category || strcmp(last_category, categories[i].name)!=0 || last_category_metric != categories[i].metric) {
		  cur_cat = ast_category_new(categories[i].name);
		  if (!cur_cat) {
		    ast_log(LOG_WARNING, "Out of memory!\n");
		    break;
		  }
		  last_category=categories[i].name;
		  last_category_metric = categories[i].metric;
		  ast_category_append(cfg, cur_cat);
		}
		new_v = ast_variable_new(categories[i].variable_name, categories[i].variable_value);
		ast_variable_append(cur_cat, new_v);
	      }
	    }
	    free(vars);
	    free(categories);
	  } 
	  else {
	    ast_log(LOG_WARNING, "LDAP RealTime: Could not find config '%s' in database.\n", file);
	  }
	}
	ast_config_log("LDAP RealTime: Configuration",cfg);
	return cfg;

}

static struct ast_config_engine ldap_engine = {
	.name = "ldap",
	.load_func = config_ldap,
	.realtime_func = realtime_ldap,
	.realtime_multi_func = realtime_multi_ldap,
	.update_func = NULL
};

int load_module (void)
{
	parse_config();

	ast_mutex_lock(&ldap_lock);

	if(!ldap_reconnect()) {
		ast_log(LOG_WARNING, "LDAP RealTime: Couldn't establish connection. Check debug.\n");
	}

	ast_config_engine_register(&ldap_engine);
	if(option_verbose) {
		ast_verbose("LDAP RealTime driver loaded.\n");
	}
	ast_cli_register(&cli_realtime_ldap_status);

	ast_mutex_unlock(&ldap_lock);

	return 0;
}

int unload_module (void)
{
	/* Aquire control before doing anything to the module itself. */
	ast_mutex_lock(&ldap_lock);

	table_configs_free();

	if (ldapConn)
	  {
	    ldap_unbind_s(ldapConn);
	    ldapConn=NULL;
	  };
	ast_cli_unregister(&cli_realtime_ldap_status);
	ast_config_engine_deregister(&ldap_engine);
	if(option_verbose) {
		ast_verbose("LDAP RealTime unloaded.\n");
	}

	STANDARD_HANGUP_LOCALUSERS;

	/* Unlock so something else can destroy the lock. */
	ast_mutex_unlock(&ldap_lock);

	return 0;
}

int reload (void)
{
	/* Aquire control before doing anything to the module itself. */
	ast_mutex_lock(&ldap_lock);

	if (ldapConn)
	  {
	    ldap_unbind_s(ldapConn);
	    ldapConn=NULL;
	  };
	parse_config();

	if(!ldap_reconnect()) {
		ast_log(LOG_WARNING, "LDAP RealTime: Couldn't establish connection. Check debug.\n");
	}

	ast_verbose(VERBOSE_PREFIX_2 "LDAP RealTime reloaded.\n");

	/* Done reloading. Release lock so others can now use driver. */
	ast_mutex_unlock(&ldap_lock);

	return 0;
}

int parse_config (void)
{
	struct ast_config *config;
	char *s;

	config = ast_config_load(RES_CONFIG_LDAP_CONF);

	if(config) {
		if(!(s=ast_variable_retrieve(config, "_general", "dbuser"))) {
			ast_log(LOG_WARNING, "LDAP RealTime: No database user found, anonymous binding as default.\n");
			dbuser[0]='\0';
		} else {
			strncpy(dbuser, s, sizeof(dbuser) - 1);
		}

		if(!(s=ast_variable_retrieve(config, "_general", "dbpass"))) {
                        ast_log(LOG_WARNING, "LDAP RealTime: No database password found, using 'asterisk' as default.\n");
                        strncpy(dbpass, "asterisk", sizeof(dbpass) - 1);
                } else {
                        strncpy(dbpass, s, sizeof(dbpass) - 1);
                }

		if(!(s=ast_variable_retrieve(config, "_general", "dbhost"))) {
                        ast_log(LOG_ERROR, "LDAP RealTime: No database host found.\n");
			dbhost[0] = '\0';
                } else {
                        strncpy(dbhost, s, sizeof(dbhost) - 1);
                }

		if(!(s=ast_variable_retrieve(config, "_general", "dbbasedn"))) {
                        ast_log(LOG_ERROR, "LDAP RealTime: No LDAP base dn found, using 'asterisk' as default.\n");
			dbbasedn[0] = '\0';
                } else {
                        strncpy(dbbasedn, s, sizeof(dbbasedn) - 1);
                }

		if(!(s=ast_variable_retrieve(config, "_general", "dbport"))) {
                        ast_log(LOG_WARNING, "LDAP RealTime: No database port found, using 389 as default.\n");
			dbport = 389;
                } else {
			dbport = atoi(s);
                }

		table_configs_free();

		char* category_name=NULL;
		while((category_name=ast_category_browse(config,category_name)))
		  {
		    ast_log(LOG_DEBUG, "LDAP RealTime: found: category_name=%s\n",category_name);
		    if (strcasecmp(category_name,"_general")!=0)
		      {
			struct ast_variable* var=ast_variable_browse(config,category_name);
			if (var)
			  {
			    struct ldap_table_config* table_config=table_config_for_table_name(category_name);			    
			    if (!table_config)
			      {
				table_config=table_config_new(category_name);
				if (table_configs)
				  table_config->next=table_configs;
				table_configs=table_config;
			      };
			    while(var)
			      {
				if (strcasecmp(var->name,"attributes")==0)
				  {				
				    ldap_table_config_add_attributes(table_config,var->value);
				  }
				else if (strcasecmp(var->name,"additionalFilter")==0)
				  {
				    table_config->additional_filter=strdup(var->value);
				  };
				var = var->next;
			      };
			  };
		      }
		  };

	}
	ast_config_destroy(config);
	/*
	ast_log(LOG_DEBUG, "LDAP RealTime Host: %s\n", dbhost);
	ast_log(LOG_DEBUG, "LDAP RealTime Port: %i\n", dbport);
	ast_log(LOG_DEBUG, "LDAP RealTime User: %s\n", dbuser);
	ast_log(LOG_DEBUG, "LDAP RealTime Password: %s\n", dbpass);
	ast_log(LOG_DEBUG, "LDAP RealTime BaseDN: %s\n", dbbasedn);
	*/
	return 1;
}

char *description (void)
{
	return res_config_ldap_desc;
}

int usecount (void)
{
	/* Try and get a lock. If unsuccessful, than that means another thread is using the ldap object. */
	if(ast_mutex_trylock(&ldap_lock)) {
		ast_log(LOG_DEBUG, "LDAP RealTime: Module usage count is 1.\n");
		return 1;
	}
	ast_mutex_unlock(&ldap_lock);
	return 0;
}

char *key ()
{
	return ASTERISK_GPL_KEY;
}

static int ldap_reconnect(void)
{
        /* mutex lock should have been locked before calling this function. */

	if (!ldapConn) {
	  if(dbhost) {
	    ldapConn=ldap_init(dbhost,dbport);
	    if (ldapConn) {
	      int bind_result=0;
	      //ldap_set_option(ldapConn, LDAP_OPT_TIMELIMIT, &timeout);
	      //ldap_set_option(ldapConn, LDAP_OPT_PROTOCOL_VERSION, &version);
	      if (dbuser && *dbuser) {
		ast_log(LOG_DEBUG, "LDAP RealTime: bind to %s as %s\n", dbhost, dbuser);
		bind_result = ldap_simple_bind_s(ldapConn, dbuser, dbpass);
	      } 
	      else {
		ast_log(LOG_DEBUG, "LDAP RealTime: bind anonymously %s anonymously\n", dbhost);
		bind_result = ldap_simple_bind_s(ldapConn, NULL, NULL);
	      }
	      if (bind_result==LDAP_SUCCESS) {
		ast_log(LOG_DEBUG, "LDAP RealTime: Successfully connected to database.\n");
		connect_time = time(NULL);
		return 1;
	      }
	      else {
		ast_log(LOG_WARNING, "LDAP RealTime: bind failed: %s\n", ldap_err2string(bind_result));
		ldap_unbind(ldapConn);
		ldapConn=NULL;
		return 0;
	      }
	    }
	    else {
	      ast_log(LOG_ERROR, "LDAP RealTime: Failed to init ldap connection to %s. Check debug for more info.\n", dbhost);
	      return 0;
	    }
	  }
	  else {
	    ast_log(LOG_ERROR, "LDAP RealTime: Not enough parameters to connect to ldap database\n");
	  }
	} 
	else {
	  ast_log(LOG_DEBUG, "LDAP RealTime: Everything seems fine.\n");
	  return 1;
	}
}

static int realtime_ldap_status(int fd, int argc, char **argv)
{
	char status[256], status2[100] = "";
	int ctime = time(NULL) - connect_time;

	if(ldapConn) {
		if(dbhost) {
			snprintf(status, 255, "Connected to %s, port %d baseDN %s", dbhost, dbport, dbbasedn);
		} 

		if(dbuser && *dbuser) {
			snprintf(status2, 99, " with username %s", dbuser);
		}

		if (ctime > 31536000) {
			ast_cli(fd, "%s%s for %d years, %d days, %d hours, %d minutes, %d seconds.\n", status, status2, ctime / 31536000, (ctime % 31536000) / 86400, (ctime % 86400) / 3600, (ctime % 3600) / 60, ctime % 60);
		} else if (ctime > 86400) {
			ast_cli(fd, "%s%s for %d days, %d hours, %d minutes, %d seconds.\n", status, status2, ctime / 86400, (ctime % 86400) / 3600, (ctime % 3600) / 60, ctime % 60);
		} else if (ctime > 3600) {
			ast_cli(fd, "%s%s for %d hours, %d minutes, %d seconds.\n", status, status2, ctime / 3600, (ctime % 3600) / 60, ctime % 60);
		} else if (ctime > 60) {
			ast_cli(fd, "%s%s for %d minutes, %d seconds.\n", status, status2, ctime / 60, ctime % 60);
		} else {
			ast_cli(fd, "%s%s for %d seconds.\n", status, status2, ctime);
		}

		return RESULT_SUCCESS;
	} else {
		return RESULT_FAILURE;
	}
}
